//******* randdev.c		by Luigi Marino, 2010 - Updated and recompiled in 2017 for Max 7 at 32 and 64 bit

#include "ext.h"							// standard Max include, always required
#include "ext_obex.h"
#define MY_RAND_MAX ((1U << 31) - 1)	

////////////////////////// object struct
typedef struct _randdev
	{
		t_object					r_ob;			
		t_atom						r_cval;
		t_atom						r_dev;
		double						r_alpha;
		double						r_trans;
		short						r_type;
		unsigned int				r_rand;		//this variable keeps track of the random series on each single instance 		
		void						*r_out;
	} t_randdev;

///////////////////////// function prototypes
//// standard set
void *randdev_new(t_symbol *s, long argc, t_atom *argv);
void randdev_assist(t_randdev *x, void *b, long m, long a, char *s);
//// additional methods
void randdev_bang(t_randdev *x);
void randdev_ft1(t_randdev *x, double f);
void randdev_ft2(t_randdev *x, double f);
void randdev_ft3(t_randdev *x, double f);
void randdev_seed(t_randdev *x, long n);
//// other functions
double deviation_float(t_randdev *x, double cval, double dev, double alpha);
long deviation_int(t_randdev *x, long cval, long dev, double alpha);
double reflect_float(double f, double min, double max);
long reflect_int(long n, long min, long max);
bool even(long n);
long myround(double f);
int myrand(t_randdev *x);
void mysrand(t_randdev *x, unsigned int i);

t_class *randdev_class;		// global pointer to the object class - so max can reference the object

//--------------------------------------------------------------------------

void ext_main(void *r)
{
	t_class *c;
	
	c = class_new("randdev", (method)randdev_new, (method)NULL, (long)sizeof(t_randdev), 0L, A_GIMME, 0);
	
    class_addmethod(c, (method)randdev_assist, "assist", A_CANT, 0);  
	class_addmethod(c, (method)randdev_bang, "bang", A_FLOAT, 0);
	class_addmethod(c, (method)randdev_ft1, "ft1", A_FLOAT, 0);
	class_addmethod(c, (method)randdev_ft2, "ft2", A_FLOAT, 0);
	class_addmethod(c, (method)randdev_ft3, "ft3", A_FLOAT, 0);
	class_addmethod(c, (method)randdev_seed, "seed", A_LONG, 0);
	
	class_register(CLASS_BOX, c);
	randdev_class = c;

}

//--------------------------------------------------------------------------

void randdev_assist(t_randdev *x, void *b, long m, long a, char *s)
{
	if (m == ASSIST_INLET) {
		switch (a) {
			case 0:
				//sprintf(s, "Inlet %ld: bang generates the value", a);
                sprintf(s, "bang generates the value");
				break;
			case 1:
				sprintf(s, "center value");
				break;
			case 2:
				sprintf(s, "deviation");
				break;
			case 3:
				sprintf(s, "distribution");
				break;
		}
	} else {	
		sprintf(s, "output");
	}
}

void randdev_bang(t_randdev *x)
{
	if (x->r_type == A_FLOAT)
		outlet_float(x->r_out, deviation_float(x, atom_getfloat(&x->r_cval), atom_getfloat(&x->r_dev), x->r_alpha));
	else {	
		outlet_int(x->r_out, deviation_int(x, atom_getlong(&x->r_cval), atom_getlong(&x->r_dev), x->r_alpha));		
	}
}

void randdev_ft1(t_randdev *x, double f)
{
	atom_setfloat(&x->r_cval, f);
}

void randdev_ft2(t_randdev *x, double f)
{
	atom_setfloat(&x->r_dev, fabs(f));
}

void randdev_ft3(t_randdev *x, double f)
{
	x->r_alpha = f;
}

void randdev_seed(t_randdev *x, long n)
{
	mysrand(x, (unsigned int)n);
}

static int makeseed(void)
{
	static unsigned int random_nextseed = 1489853723;
	random_nextseed = random_nextseed * 435898247 + 938284287;
	return (random_nextseed & 0x7fffffff);
}		

void mysrand(t_randdev *x, unsigned int n)  
{	
	if (n == 0) {
		x->r_rand = makeseed();  
	} else
		x->r_rand = n;
	/*
	 In order to initialize randomly the series, it is usually used the system clock as seed. 
	 Doing so, if you have multiple instances of a random generator on the same patch they'll be initialized at the same time getting the same seed.
	 makeseed() is declared as static so that when it is called by multiple instances at the same time you get different seeds.
	 */
}

int myrand(t_randdev *x) //Linear Congruential Generator (LCG) 
{
	return (x->r_rand = (x->r_rand * 1103515245 + 12345) & MY_RAND_MAX);
	/*
	 Implementation of rand() rewritten using a variable in the object structure to allow a separate random series for each object instance. 
	 The quality of a LCG depends a lot on the multiplier and increment (here 1103515245 and 12345).
	 For other methods to generate random numbers check out Mersenne Twister or R250/521.  
	 */						
}


double deviation_float(t_randdev *x, double cval, double dev, double alpha)
{
	double out;
	double range;
	double uni, cau;
	const double PI = 3.14159265358979; 
	const double cf = 0.77;  //compensation factor, it's useful... maybe... 
	
	range = fabs(dev * 2.0);
	
	if (alpha < 0.) {
		alpha = fabs(alpha);
		do
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while (uni == 0.5);
		if (dev >= 0.0) {
			cau = alpha * (tan(PI * uni));
			out = (cau * (range * cf)) + cval;
		} else {
			cau = alpha * (tan(PI * uni));
			out = (cau * (range * cf)) + cval;
		}
	} else if (alpha == 0.0) { 
		out = cval;	
	} else if (alpha < 1.0) {
		do
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while (uni == 0.5);
		if (dev >= 0.0) {
			cau = alpha * (tan(PI * uni));
			out = reflect_float((cau * (range * cf)), -dev, dev) + cval;
		} else {
			cau = alpha * (tan(PI * uni));
			out = (cau * (range * cf)) + cval;
		}
	} else {
		uni = (double)myrand(x) / MY_RAND_MAX;
		out = (uni * range) + (cval - dev); 
	}

	return out;
}

long deviation_int(t_randdev *x, long cvallong, long devlong, double alpha)
{
	long out;
	double range, cval, dev;
	double uni, cau;
	const double PI = 3.14159265358979; 
	const double cf = 0.77;
	
	cval = (double)cvallong;
	dev = (double)devlong;

	range = fabs(dev * 2) + 1;

	if (alpha < 0.) {
		alpha = fabs(alpha);	
		do 
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while ((uni == 0.5) || (uni == 1.0));
		if (dev >= 0) { 
			cau = alpha * (tan(PI * uni)) + 0.5;
			out = myround((cau * range) + (cval - (range / 2.0)));
		} else {
			cau = alpha * tan(PI * uni);
			out = myround(cau * (range * cf)) + cval;
		}			
	} else if (alpha == 0.0) {
		out = cval;	
	} else if (alpha < 1.0) {
		do 
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while ((uni == 0.5) || (uni == 1.0));
		if (dev >= 0) { 
			cau = reflect_float(alpha * (tan(PI * uni)), -0.5, 0.5) + 0.5;
			out = myround((cau * range) + (cval - (range / 2.0)));
		} else {
			cau = alpha * tan(PI * uni);
			out = myround(cau * (range * cf)) + cval;
		}	
	} else {
		do 
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while (uni == 1.0);
		out = ((long)(uni * range) - dev) + cval; 
	}		

	return out;
}
	
double reflect_float(double f, double min, double max)
{
	double rng, frel;
	unsigned long c;
	
	rng = fabs(max - min);
	
	if (rng == 0) 
		return min;
	
	frel = f - min;
	c = (long)(frel / rng);
	
	if (!even(c)) 
		frel = rng - fabs(fmod(frel, rng));
	else
		frel = fabs(fmod(frel, rng));
	
	f = frel + min;
	
	return f;
}	

long reflect_int(long n, long min, long max)
{
	long rng, nrel;
	unsigned long c;
	
	rng = abs(max - min);
	
	if (rng == 0)
		return min;
	
	nrel = n - min;
	c = (unsigned long)(nrel / rng);
	
	if (!even(c)) 
		nrel = rng - abs(nrel % rng);
	else
		nrel = abs(nrel % rng);
	
	n = nrel + min;
	
	return n;
}	

bool even(long n)
{
	if ((n % 2) == 0) 
		return true;
	else 
		return false;
}

long myround(double f)
{
	long n;
	if (f >= 0) {
		if (fmod(f, 1.0) >= 0.5)
			n = (long)f + 1;
		else
			n = (long)f;
	} else {
		if (fmod(f, 1.0) <= -0.5)
			n = (long)f - 1;
		else
			n = (long)f;
	}
	return n;
}	


//--------------------------------------------------------------------------


void *randdev_new(t_symbol *s, long argc, t_atom *argv)
{
    t_randdev *x;
    
    x = (t_randdev *)object_alloc(randdev_class);
    
    switch (argc) {
        case 0:
            x->r_type = A_FLOAT;
            atom_setfloat(&x->r_cval, 0.0);
            atom_setfloat(&x->r_dev, -1.0);
            x->r_alpha = 0.0;
            break;
        case 1:
            if (atom_gettype(argv) == A_FLOAT) {
                x->r_type = A_FLOAT;
                atom_setfloat(&x->r_cval, 0.0);
                atom_setfloat(&x->r_dev, -1.0);
            } else {
                x->r_type = A_LONG;
                atom_setlong(&x->r_cval, 0);
                atom_setlong(&x->r_dev, 50);
            }
            x->r_alpha = 0.0;
            break;
        case 3:
            if ((atom_gettype(argv) == A_FLOAT) || (atom_gettype(argv+1) == A_FLOAT)) {
                x->r_type = A_FLOAT;
                atom_setfloat(&x->r_cval, atom_getfloat(argv));
                atom_setfloat(&x->r_dev, atom_getfloat(argv+1));
            } else {
                x->r_type = A_LONG;
                atom_setlong(&x->r_cval, atom_getlong(argv));
                atom_setlong(&x->r_dev, atom_getlong(argv+1));
            }
            x->r_alpha = atom_getfloat(argv+2);
            break;
        default:
            object_warn((t_object *)x, "wrong arguments, default initialization");
            x->r_type = A_FLOAT;
            atom_setfloat(&x->r_cval, 0.0);
            atom_setfloat(&x->r_dev, -1.0);
            x->r_alpha = 0.0;
            break;
    }
    
    mysrand(x, 0);
    
    floatin(x,3);
    floatin(x,2);
    floatin(x,1);
    
    x->r_out = outlet_new(x, NULL);
    
    return(x);
}

